import torch
from torch import nn
from torch.nn import init


class SlotAttention(nn.Module):
    """
    Implementation of Slot Attention from the paper 'Object-Centric Learning
    with Slot Attention' in Pytorch by lucidrains.
    https://github.com/lucidrains/slot-attention/blob/master/slot_attention/slot_attention.py
    
    Slightly modified to allow turning off layer norm. 
    
    Example
    -------
    >>> B, Din = 8, 64
    >>> N, Dout = 4, 32
    >>> inputs = torch.rand([B, N, Din])
    >>> net = SlotAttentionOriginal(dim=Dout, kvdim=Din)
    >>> outputs = net(inputs)
    >>> outputs.shape == torch.Size([B, N, Dout])
    True
    """
    def __init__(self, num_slots, dim, iters = 3, eps = 1e-8, hidden_dim = 128, use_ln=True):
        super().__init__()
        self.num_slots = num_slots
        self.iters = iters
        self.eps = eps
        self.scale = dim ** -0.5
        self.use_ln = use_ln

        self.slots_mu = nn.Parameter(torch.randn(1, 1, dim))

        self.slots_logsigma = nn.Parameter(torch.zeros(1, 1, dim))
        init.xavier_uniform_(self.slots_logsigma)

        self.to_q = nn.Linear(dim, dim)
        self.to_k = nn.Linear(dim, dim)
        self.to_v = nn.Linear(dim, dim)

        self.gru = nn.GRUCell(dim, dim)

        hidden_dim = max(dim, hidden_dim)

        self.mlp = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.ReLU(inplace = True),
            nn.Linear(hidden_dim, dim)
        )
        if self.use_ln:
            self.norm_input  = nn.LayerNorm(dim)
            self.norm_slots  = nn.LayerNorm(dim)
            self.norm_pre_ff = nn.LayerNorm(dim)

    def forward(self, inputs, slots=None, num_slots = None):
        b, n, d, device = *inputs.shape, inputs.device
        n_s = num_slots if num_slots is not None else self.num_slots
        
        if slots is None:
            mu = self.slots_mu.expand(b, n_s, -1)
            sigma = self.slots_logsigma.exp().expand(b, n_s, -1)

            slots = mu + sigma * torch.randn(mu.shape, device = device)

        if self.use_ln:
            inputs = self.norm_input(inputs)        
        k, v = self.to_k(inputs), self.to_v(inputs)

        for _ in range(self.iters):
            slots_prev = slots

            if self.use_ln:
                slots = self.norm_slots(slots)
            q = self.to_q(slots)

            dots = torch.einsum('bid,bjd->bij', q, k) * self.scale
            attn = dots.softmax(dim=1) + self.eps
            attn = attn / attn.sum(dim=-1, keepdim=True)

            updates = torch.einsum('bjd,bij->bid', v, attn)

            slots = self.gru(
                updates.reshape(-1, d),
                slots_prev.reshape(-1, d)
            )

            slots = slots.reshape(b, -1, d)
            
            if not self.use_ln:
                slots = slots + self.mlp(slots)
            else:
                slots = slots + self.mlp(self.norm_pre_ff(slots))

        return slots
